/***************************************************************************//*!
\file bdd.h
\brief File bdd.h contains declaration of all external data structures.

### Description

    PackageName [bdd]
    Synopsis    [Package 'bdd' enables symbolic computations with BDDs.]

    FileName    [bdd.h]
    Revision    [$Revision: 188 $]
    Date        [$Date: 2019-02-07 15:11:19 +0100 (čet, 07 feb 2019) $]
    Authors     [Robert Meolic (robert@meolic.com),
                 Ales Casar (ales@homemade.net)]

### Copyright

This file is part of EST (Efficient Symbolic Tools).
Copyright (C) 2003, 2019 UM FERI, Koroska cesta 46, SI-2000 Maribor, Slovenia

EST is free software; you can redistribute it and/or modify it under the terms
of the GNU General Public License as published by the Free Software Foundation;
either version 2 of the License, or (at your option) any later version.

EST is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
PURPOSE. See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License along with
this program; if not, write to the Free Software Foundation, Inc., 51 Franklin
Street, Fifth Floor, Boston, MA 02110-1301 USA.

### More info

See also: bddInt.h

*******************************************************************************/

#ifndef _BDD
#define _BDD

#include <est.h>

/* ON MS WINDOWS THERE HAS TO BE DEFINED ESTWIN32 */
/* ON GNU/LINUX THERE HAS TO BE DEFINED ESTUNIX */
/* ON MACOSX THERE HAS TO BE DEFINED ESTMACOSX */

#ifdef ESTUNIX
#  ifndef EXTERN
#    define EXTERN extern
#  endif
#  ifndef DATAEXTERN
#    define DATAEXTERN extern
#  endif
#endif

#ifdef ESTMACOSX
#  ifndef EXTERN
#    define EXTERN extern
#  endif
#  ifndef DATAEXTERN
#    define DATAEXTERN extern
#  endif
#endif

#ifdef ESTWIN32
#  ifdef TCL
#    undef EXTERN
#    define EXTERN extern
#    undef DATAEXTERN
#    define DATAEXTERN extern
#  else
#    if defined(BUILD_EST) || defined(BUILD_BDD)
#      undef EXTERN
#      define EXTERN __declspec (dllexport)
#      undef DATAEXTERN
#      define DATAEXTERN extern __declspec (dllexport)
#    else
#      if defined(USE_EST) || defined(USE_BDD)
#        undef EXTERN
#        define EXTERN __declspec (dllimport)
#        undef DATAEXTERN
#        define DATAEXTERN __declspec (dllimport)
#      else
#        undef EXTERN
#        define EXTERN extern
#        undef DATAEXTERN
#        define DATAEXTERN extern
#      endif
#    endif
#  endif
#endif

/*----------------------------------------------------------------------------*/
/* Constant definitions                                                       */
/*----------------------------------------------------------------------------*/

/*----------------------------------------------------------------------------*/
/* Macro definitions                                                          */
/*----------------------------------------------------------------------------*/

/* BIDDY */
/**/
#define Bdd_isEqv(f,g) ((f).p == (g).p)
#define Bdd_isConstant(f) (((f).p == bdd_termTrue.p || (f).p == bdd_termFalse.p))
#define Bdd_isNull(f) ((f).p == bdd_termNull.p)
/**/

/*----------------------------------------------------------------------------*/
/* Type declarations                                                          */
/*----------------------------------------------------------------------------*/

typedef unsigned int Bdd_Variable; /* must be compatible with Biddy_Variable */

typedef struct {
  void *p; /* pointer into node table */
  Est_Boolean mark; /* mark */
} Bdd_Edge;

/* currently, not used */
/*
typedef void (*Bdd_GCFunction)(Biddy_Manager);
*/

/*----------------------------------------------------------------------------*/
/* Variable declarations                                                      */
/*----------------------------------------------------------------------------*/

#ifdef __cplusplus
extern "C" {
#endif

DATAEXTERN int bdd_status;

DATAEXTERN Bdd_Edge bdd_termTrue; /* terminal node */

DATAEXTERN Bdd_Edge bdd_termFalse; /* terminal node */

DATAEXTERN Bdd_Edge bdd_termNull; /* null edge */

#ifdef __cplusplus
}
#endif

/*----------------------------------------------------------------------------*/
/* Prototypes for functions exported from bddMain.c                           */
/*----------------------------------------------------------------------------*/

#ifdef __cplusplus
extern "C" {
#endif

EXTERN void Bdd_InitPkg();

EXTERN void Bdd_ExitPkg();

EXTERN void Bdd_AboutPkg();

EXTERN void Bdd_SetManagerParameters(float gcr, float gcrF, float gcrX,
            float rr, float rrF, float rrX, float st, float cst);

EXTERN unsigned int Bdd_VariableNum();

EXTERN Bdd_Variable Bdd_FoaVariable(Est_String x);

EXTERN Est_String Bdd_GetVariableName(Bdd_Variable v);

EXTERN Bdd_Edge Bdd_GetVariableEdge(Bdd_Variable v);

EXTERN void Bdd_SetVariableWeight(Bdd_Variable v, unsigned int w);

EXTERN unsigned int Bdd_GetVariableWeight(Bdd_Variable v);

EXTERN void Bdd_SetVariableProcComp(Bdd_Variable v, int n);

EXTERN int Bdd_GetVariableProcComp(Bdd_Variable v);

EXTERN Est_Boolean Bdd_IsSmaller(Bdd_Edge f, Bdd_Edge g);

EXTERN Est_Boolean Bdd_IsVariableSmaller(Bdd_Variable fv, Bdd_Variable gv);

EXTERN void Bdd_ResetVariablesValue();

EXTERN void Bdd_SetVariableValue(Bdd_Variable v, Bdd_Edge f);

EXTERN Bdd_Edge Bdd_NOT(Bdd_Edge f);

EXTERN Bdd_Edge Bdd_AND(Bdd_Edge f, Bdd_Edge g);

EXTERN Bdd_Edge Bdd_OR(Bdd_Edge f, Bdd_Edge g);

EXTERN Bdd_Edge Bdd_XOR(Bdd_Edge f, Bdd_Edge g);

EXTERN Bdd_Edge Bdd_ITE(Bdd_Edge f, Bdd_Edge g, Bdd_Edge h);

EXTERN Bdd_Edge Bdd_VarITE(Bdd_Variable v, Bdd_Edge g, Bdd_Edge h);

EXTERN Bdd_Edge Bdd_Restrict(Bdd_Edge f, Bdd_Variable v, Est_Boolean value);

EXTERN Bdd_Edge Bdd_Compose(Bdd_Edge f, Bdd_Variable v, Bdd_Edge g);

EXTERN Bdd_Edge Bdd_E(Bdd_Edge f, Bdd_Variable v);

EXTERN Bdd_Edge Bdd_A(Bdd_Edge f, Bdd_Variable v);

EXTERN Bdd_Edge Bdd_Support(Bdd_Edge f);

EXTERN Est_String Bdd_PopFromSupport(Bdd_Edge *spp);

EXTERN Bdd_Edge Bdd_Replace(Bdd_Edge f, Est_String keyword);

EXTERN double Bdd_CountMinterm(Bdd_Edge f, unsigned int nvars);

EXTERN Bdd_Edge Bdd_Eval1(Est_String s);

EXTERN Bdd_Edge Bdd_Eval2(Est_String s);

EXTERN void Bdd_Clean();

EXTERN void Bdd_Garbage();

EXTERN void Bdd_Refresh(Bdd_Edge f);

EXTERN void Bdd_Reorder(Bdd_Edge f);

EXTERN Est_Boolean Bdd_FindFormula(Est_String x, Bdd_Edge *f);

EXTERN void Bdd_SaveFormula(Est_String prefix, Est_String name, Bdd_Edge f);

EXTERN void Bdd_SaveTmpFormula(Bdd_Edge f, int count);

EXTERN void Bdd_DeleteFormula(Est_String prefix, Est_String name);

EXTERN void Bdd_SystemStat(FILE *s);

EXTERN void Bdd_FunctionStat(FILE *s, Bdd_Edge f);

EXTERN unsigned long long int Bdd_Stat_ReadMemoryInUse();

EXTERN unsigned int Bdd_Stat_NodeTableSize();

EXTERN unsigned int Bdd_Stat_NodeTableMax();

EXTERN unsigned int Bdd_Stat_NodeTableGenerated();

EXTERN unsigned int Bdd_Stat_NodeTableGCNumber();

#ifdef __cplusplus
}
#endif

/*----------------------------------------------------------------------------*/
/* Prototypes for functions exported from bddOutput.c                         */
/*----------------------------------------------------------------------------*/

#ifdef __cplusplus
extern "C" {
#endif

EXTERN void Bdd_WriteBDD(Est_String name);

EXTERN void Bdd_WriteTable(Est_String name);

EXTERN void Bdd_WriteFunction(Est_String name);

EXTERN int Bdd_WriteDot(Est_String filename, Est_String name, int id);

EXTERN int Bdd_WriteBDDView(Est_String filename, Est_String name, int id);

#ifdef __cplusplus
}
#endif

/*----------------------------------------------------------------------------*/
/* Prototypes for functions exported from bddUtil.c                           */
/*----------------------------------------------------------------------------*/

#ifdef __cplusplus
extern "C" {
#endif

EXTERN Bdd_Edge Bdd_EncodeMinterm(Bdd_Variable *tabVar, int n, int i, int min);

EXTERN Bdd_Edge Bdd_ExtractBranch(Bdd_Edge f);

EXTERN Bdd_Edge Bdd_Expand(Est_String fname, Bdd_Edge f);

EXTERN void Bdd_DecodeA(Bdd_Edge f, Est_Boolean mark, unsigned int minA,
            unsigned int maxA, unsigned int *rnum, unsigned int **rmin,
            unsigned int **rmax);

EXTERN void Bdd_DecodeR(Bdd_Edge f, Est_Boolean mark, unsigned int minR,
            unsigned int maxR, unsigned int *rnum, unsigned int **rmin,
            unsigned int **rmax);

EXTERN void Bdd_DecodeRAS(Bdd_Edge f, Est_Boolean mark, unsigned int minR,
            unsigned int maxR, unsigned int minA, unsigned int maxA,
            unsigned int minS, unsigned int maxS, unsigned int *rnum,
            unsigned int **rminr, unsigned int **rmaxr,
            unsigned int **rmina, unsigned int **rmaxa,
            unsigned int **rmins, unsigned int **rmaxs);

EXTERN void Bdd_DecodeRRR(unsigned int n, Bdd_Edge f, Est_Boolean mark,
            unsigned int *decodingTableR,
            unsigned int *rnum, unsigned int **rmin, unsigned int **rmax);

EXTERN void Bdd_DecodeRRRASSS(unsigned int n, Bdd_Edge f, Est_Boolean mark,
            unsigned int minA, unsigned int maxA, unsigned int *decodingTableR,
            unsigned int *decodingTableS, unsigned int *rnum,
            unsigned int **rminr, unsigned int **rmaxr,
            unsigned int **rmina, unsigned int **rmaxa,
            unsigned int **rmins, unsigned int **rmaxs);

EXTERN void Bdd_DecodeRP(Bdd_Edge f, Est_Boolean mark, unsigned int minR,
            unsigned int maxR, unsigned int minP, unsigned int maxP,
            unsigned int *rnum, unsigned int **rminr, unsigned int **rmaxr,
            unsigned int **rminp, unsigned int **rmaxp);

EXTERN void Bdd_DecodeRRRP(unsigned int n, Bdd_Edge f, Est_Boolean mark,
            unsigned int *decodingTableR,
            unsigned int minP, unsigned int maxP, unsigned int *rnum,
            unsigned int **rminr, unsigned int **rmaxr,
            unsigned int **rminp, unsigned int **rmaxp);

EXTERN void Bdd_DecodeRRRPPP(unsigned int n1, unsigned int n2, Bdd_Edge f,
            Est_Boolean mark, unsigned int *decodingTableR,
            unsigned int *decodingTableP, unsigned int *rnum,
            unsigned int **rminr, unsigned int **rmaxr,
            unsigned int **rminp, unsigned int **rmaxp);

EXTERN Est_Boolean Bdd_ExtractA(Bdd_Edge f, Est_Boolean mark,
                   unsigned int *minA);

EXTERN Est_Boolean Bdd_ExtractR(Bdd_Edge f, Est_Boolean mark,
                   unsigned int *minR);

EXTERN Est_Boolean Bdd_ExtractRAS(Bdd_Edge f, Est_Boolean mark,
                   unsigned int *minR, unsigned int *minA, unsigned int *minS);

EXTERN Est_Boolean Bdd_ExtractRRR(Bdd_Edge f, Est_Boolean mark,
                   unsigned int *decodingTableR);

EXTERN Est_Boolean Bdd_ExtractRRRASSS(Bdd_Edge f, Est_Boolean mark,
                   unsigned int *minA, unsigned int *decodingTableR,
                   unsigned int *decodingTableS);
                   
#ifdef __cplusplus
}
#endif

/*----------------------------------------------------------------------------*/
/* Prototypes for functions exported from bddRel.c                            */
/*----------------------------------------------------------------------------*/

#ifdef __cplusplus
extern "C" {
#endif

EXTERN Bdd_Edge Bdd_RelOp(Bdd_Edge f, Bdd_Edge g,
                          Est_String format, Est_Boolean firsttime);

EXTERN Bdd_Edge Bdd_RelOpSimple(Bdd_Edge f,
                                 Est_String format, Est_Boolean firsttime);

EXTERN Bdd_Edge Bdd_RelOpComplex(Bdd_Edge f,
                                 Est_String format, Est_Boolean firsttime);

EXTERN Bdd_Edge Bdd_NewState(Bdd_Edge f);

EXTERN Bdd_Edge Bdd_NewState1(Bdd_Edge f, Est_String name);

EXTERN Bdd_Edge Bdd_NewState2(
                  Bdd_Edge f, Est_String name1, Est_String name2);

EXTERN Bdd_Edge Bdd_NewEq(Bdd_Edge f, Bdd_Edge g);

EXTERN Bdd_Edge Bdd_StabRS(Bdd_Edge f, Bdd_Edge stablist);

#ifdef __cplusplus
}
#endif

#endif  /* _BDD */
