/***************************************************************************//*!
\file pa.h
\brief File pa.h contains declaration of all external data structures.

### Description

    PackageName [pa]
    Synopsis    [Package 'pa' is a framework for implementation of
                 process algebrae similar to Milner's CCS and Hoare's CSP.]

    FileName    [pa.h]
    Revision    [$Revision: 188 $]
    Date        [$Date: 2019-02-07 15:11:19 +0100 (čet, 07 feb 2019) $]
    Authors     [Robert Meolic (robert@meolic.com)]

### Copyright

This file is part of EST (Efficient Symbolic Tools).
Copyright (C) 2003, 2019 UM FERI, Koroska cesta 46, SI-2000 Maribor, Slovenia

EST is free software; you can redistribute it and/or modify it under the terms
of the GNU General Public License as published by the Free Software Foundation;
either version 2 of the License, or (at your option) any later version.

EST is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
PURPOSE. See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License along with
this program; if not, write to the Free Software Foundation, Inc., 51 Franklin
Street, Fifth Floor, Boston, MA 02110-1301 USA.

### More info

See also: paInt.h

*******************************************************************************/

#ifndef _PA
#define _PA

#include <est.h>
#include <bdd.h>

/* ON WIN32 THERE HAS TO BE DEFINED ESTWIN32 */
/* ON UNIX THERE HAS TO BE DEFINED ESTUNIX */
/* ON MACOSX THERE HAS TO BE DEFINED ESTMACOSX */

#ifdef ESTUNIX
#  ifndef EXTERN
#    define EXTERN extern
#  endif
#  ifndef DATAEXTERN
#    define DATAEXTERN extern
#  endif
#endif

#ifdef ESTMACOSX
#  ifndef EXTERN
#    define EXTERN extern
#  endif
#  ifndef DATAEXTERN
#    define DATAEXTERN extern
#  endif
#endif

#ifdef ESTWIN32
#  ifdef TCL
#    undef EXTERN
#    define EXTERN extern
#    undef DATAEXTERN
#    define DATAEXTERN extern
#  else
#    if defined(BUILD_EST) || defined(BUILD_PA)
#      undef EXTERN
#      define EXTERN __declspec (dllexport)
#      undef DATAEXTERN
#      define DATAEXTERN extern __declspec (dllexport)
#    else
#      if defined(USE_EST) || defined(USE_PA)
#        undef EXTERN
#        define EXTERN __declspec (dllimport)
#        undef DATAEXTERN
#        define DATAEXTERN __declspec (dllimport)
#      else
#        undef EXTERN
#        define EXTERN extern
#        undef DATAEXTERN
#        define DATAEXTERN extern
#      endif
#    endif
#  endif
#endif

/*----------------------------------------------------------------------------*/
/* Constant definitions                                                       */
/*----------------------------------------------------------------------------*/

/*----------------------------------------------------------------------------*/
/* Macro definitions                                                          */
/*----------------------------------------------------------------------------*/

/*----------------------------------------------------------------------------*/
/* Type declarations                                                          */
/*----------------------------------------------------------------------------*/

typedef struct {
  Est_String  name; /* action name */
  Bdd_Edge p; /* BDD representation of an action */
} Pa_Action;

typedef struct {
  Est_String name; /* state name */
  Bdd_Edge bddR ; /* BDD with 'r' variables (current state) */
  Bdd_Edge bddS; /* BDD with 's' variables (next state) */
  Bdd_Edge bddComp; /* represents how the state is composed */
  Bdd_Edge acc; /* acceptance set */
  Bdd_Edge minacc; /* minimized acceptance set */
  Est_String ccs; /* outgoing transition in CCS format */
  Est_Boolean final; /* TRUE for final states, used for w/c automata */
} Pa_State;

typedef struct {
  int source; /* source state */
  int action; /* action */
  int type; /* action type, 0 = output, 1 = input */
  int dest; /* destination state */
} Pa_Transition;

typedef struct {
  Est_String name; /* sort name */
  int encoded; /* 1 = yes, 0 = no */
  int numVar; /* number of variables */
  unsigned int maxWeight; /* weight = 1,2,4... (decoding ) */
  int numActions; /* number of actions */
  Pa_Action *table; /* table of actions */
  Bdd_Variable a0; /* first variable, i.e. a0 */
  Bdd_Variable b0; /* first variable renamed to b0 */
  Est_Boolean deleted; /* TRUE = deleted */
} Pa_Sort;

typedef struct {
  Est_String name; /* process name */
  Est_String compname; /* optionally, if states are composed */
  int sort; /* sort */
  int encoded; /* 1 = yes, 0 = no */
  int numVar; /* number of variables */
  unsigned int maxWeight; /* weight = 1,2,4... (decoding ) */
  int numStates; /* number of states */
  Pa_State *tableS; /* table of states */
  int initial; /* initial state */
  Bdd_Edge stateBDD; /* BDD of all states */
  int numTransitions; /* number of transitions */
  Pa_Transition *tableT; /* table of transitions */
  Bdd_Edge d; /* transition relation */
  Bdd_Edge stab; /* stab BDD for process */
  Bdd_Edge stablist; /* stablist BDD for process */
  Est_Boolean trace; /* TRUE = use trace minimized process during composition */
  Est_Boolean weak; /* TRUE = use weak minimized process during composition */
  Est_Boolean strong; /* TRUE = use strong minimized process during composition */
  Est_Boolean deleted; /* TRUE = deleted */
} Pa_Process;

typedef struct {
  Est_String name; /* composition name */
  int type; /* type of composition */
  int numProcesses; /* number of processes */
  int *tableP; /* table of processes */
  int numActions; /* number of given actions */
  int *tableA; /* table of actions */
  int sort; /* sort */
  Bdd_Edge initialBDD; /* BDD of initial state */
  Bdd_Edge stateBDD; /* BDD of all states */
  Bdd_Edge deadlockedBDD; /* BDD of all deadlocked states */
  Bdd_Edge transitionBDD; /* BDD of transition relation */
  Bdd_Edge actionsBDD; /* BDD of actions given to the composition */
  Bdd_Edge *di; /* di BDDs for processes */
  Bdd_Edge *dij; /* dij BDDs for processes */
  Bdd_Edge *alphabet; /* alphabets of processes */
  Bdd_Edge *dinew; /* used to improve step-by-step composition */
  Bdd_Edge *dijnew; /* used to improve step-by-step composition */
  Bdd_Edge *dicube; /* used to improve step-by-step composition */
  Bdd_Edge *dijcube; /* used to improve step-by-step composition */
  Bdd_Edge snew; /* used to improve step-by-step composition */
  Est_Boolean onthefly; /* TRUE = used for on-the-fly algorithms */
  Est_Boolean full; /* TRUE = used in CCS parser */
  Est_Boolean trace; /* TRUE = used in CCS parser */
  Est_Boolean weak; /* TRUE = used in CCS parser */
  Est_Boolean strong; /* TRUE = used in CCS parser */
} Pa_Composition;

/*----------------------------------------------------------------------------*/
/* Variable declarations                                                      */
/*----------------------------------------------------------------------------*/

#ifdef __cplusplus
extern "C" {
#endif

DATAEXTERN int pa_status;

DATAEXTERN int pa_sorts;                         /* number of sorts */

DATAEXTERN int pa_processes;                     /* number of processes */

DATAEXTERN int pa_compositions;                  /* number of compositions */

DATAEXTERN Pa_Sort *pa_sortTable;                /* table of sorts */

DATAEXTERN Pa_Process *pa_processTable;          /* table of processes */

DATAEXTERN Pa_Composition *pa_compositionTable;  /* table of compositions */

#ifdef __cplusplus
}
#endif

/*----------------------------------------------------------------------------*/
/* Prototypes for functions exported from paMain.c                            */
/*----------------------------------------------------------------------------*/

#ifdef __cplusplus
extern "C" {
#endif

EXTERN void Pa_InitPkg();

EXTERN void Pa_ExitPkg();

EXTERN void Pa_AboutPkg();

EXTERN int Pa_AddNewSort(Est_String name);

EXTERN int Pa_FindSort(Est_String name);

EXTERN int Pa_FOASortAction(Pa_Sort *sort, Est_String name);

EXTERN int Pa_FindSortAction(Pa_Sort *sort, Est_String name);

EXTERN int Pa_AddNewProcess(Est_String name);

EXTERN int Pa_FindProcess(Est_String name);

EXTERN void Pa_RenameProcess(Est_String old, Est_String new);

EXTERN int Pa_CopyProcess(Est_String name, Est_String copy,
                Est_String copysort, Est_String old, Est_String new);

EXTERN Est_Boolean Pa_IsGate(Est_String name);

EXTERN int Pa_FOAStateProcess(Pa_Process *proc, Est_String name);

EXTERN int Pa_FOAGateProcess(Pa_Process *proc, Est_String name);

EXTERN int Pa_FindStateProcess(Pa_Process *proc, Est_String name);

EXTERN int Pa_AddStateProcess(Pa_Process *proc, Est_String name);

EXTERN int Pa_AddAdaptedStateProcess(Pa_Process *proc, Est_String name);

EXTERN int Pa_AddNextStateProcess(Pa_Process *proc, Est_String prefix);

EXTERN int Pa_FindComposedStateProcess(Pa_Process *proc, Bdd_Edge F);

EXTERN int Pa_AddComposedStateProcess(Pa_Process *proc, Bdd_Edge F,
                                      Est_String prefix);

EXTERN int Pa_NumberStateProcess(Pa_Process *proc);

EXTERN int Pa_FOATransition(Pa_Process *proc, int state1,
                            int action, int type, int state2);

EXTERN int Pa_AddNewComposition(Est_String name, int type,
                                int numP, int numA);

EXTERN void Pa_DeleteStateProcess(Pa_Process *proc, int num);

EXTERN void Pa_DeleteSort(int num);

EXTERN void Pa_RemoveDeletedSorts();

EXTERN void Pa_DeleteProcess(int num);

EXTERN void Pa_RemoveDeletedProcesses();

EXTERN void Pa_RemoveComposition(int num);

EXTERN int Pa_FindComposition(Est_String name);

EXTERN int Pa_FindProcessComp(Pa_Composition *c, Est_String name);

EXTERN Bdd_Edge Pa_AddStab(Bdd_Edge d, Est_String name);

EXTERN void Pa_SaveSortActions(Pa_Sort *s, Est_Boolean full);

EXTERN void Pa_SaveProcStates(Pa_Process *p, Est_Boolean full);

#ifdef __cplusplus
}
#endif

/*----------------------------------------------------------------------------*/
/* Prototypes for functions exported from paKod.c                             */
/*----------------------------------------------------------------------------*/

#ifdef __cplusplus
extern "C" {
#endif

EXTERN void Pa_EncodeSort(Est_String name);

EXTERN void Pa_EncodeProcess(Est_String name);

EXTERN void Pa_EncodeMultiProcesses(Est_String plist, Est_String sort);

EXTERN int Pa_ExpandGates(Pa_Process *p);

EXTERN int Pa_Process2Composition(Est_String name);

EXTERN int Pa_Composition2Process(Est_String name, Est_Boolean full);

EXTERN unsigned int Pa_DecodeSortActions(Pa_Sort *s, Bdd_Edge F, Est_Boolean write);

EXTERN Est_String Pa_DecodeSortActions1(Pa_Sort *s, Bdd_Edge F, Est_Boolean write);

EXTERN void Pa_DecodeSort(Est_String name);

EXTERN void Pa_DecodeProcess(Est_String name, Est_Boolean write);

EXTERN unsigned int Pa_DecodeProcessStates(Pa_Process *p, Bdd_Edge F,
                                   Est_Boolean write);

EXTERN unsigned long long int Pa_DecodeProcessTR(Pa_Process *p, Bdd_Edge F,
                               Est_Boolean write);

EXTERN void Pa_DecodeComposition(Est_String name, Est_Boolean write);

EXTERN unsigned int Pa_DecodeCompStates(Pa_Composition *c, Bdd_Edge F,
                                Est_Boolean write);

EXTERN unsigned long long int Pa_DecodeCompTR(Pa_Composition *c, Bdd_Edge F,
                                Est_Boolean write);

EXTERN unsigned int Pa_DecodeProcProcPair(Pa_Process *p1, Pa_Process *p2,
                                Bdd_Edge F, Est_Boolean write);

EXTERN unsigned int Pa_DecodeCompProcPair(Pa_Composition *c, Pa_Process *p,
                                Bdd_Edge F, Est_Boolean write);

EXTERN unsigned int Pa_DecodeCompCompPair(Pa_Composition *c1, Pa_Composition *c2,
                                Bdd_Edge F, Est_Boolean write);

EXTERN Bdd_Edge Pa_ExtractSortAction(Pa_Sort *s, Bdd_Edge f);

EXTERN Bdd_Edge Pa_ExtractProcState(Pa_Process *p, Bdd_Edge f);

EXTERN Bdd_Edge Pa_ExtractProcTransition(Pa_Process *p, Bdd_Edge f);

EXTERN Bdd_Edge Pa_ExtractCompState(Pa_Composition *c, Bdd_Edge f);

EXTERN Bdd_Edge Pa_ExtractCompTransition(Pa_Composition *c, Bdd_Edge f);

EXTERN unsigned int Pa_EnumerateSortAction(Pa_Sort *s, Bdd_Edge f);

EXTERN unsigned int Pa_EnumerateProcState(Pa_Process *p, Bdd_Edge f);

#ifdef __cplusplus
}
#endif

/*----------------------------------------------------------------------------*/
/* Prototypes for functions exported from paOutput.c                          */
/*----------------------------------------------------------------------------*/

#ifdef __cplusplus
extern "C" {
#endif

EXTERN void Pa_WriteKnownSorts();

EXTERN void Pa_WriteSort(Est_String name);

EXTERN void Pa_WriteKnownProcesses();

EXTERN void Pa_WriteProcess(Est_String name, char type);

EXTERN void Pa_WriteKnownCompositions();

EXTERN void Pa_WriteProcessCCS(Est_String name, char type);

EXTERN void Pa_WriteProcessFSP(Est_String name, char type);

#ifdef __cplusplus
}
#endif

/*----------------------------------------------------------------------------*/
/* Prototypes for functions exported from paParser.y                          */
/*----------------------------------------------------------------------------*/

#ifdef __cplusplus
extern "C" {
#endif

EXTERN int Pa_ParseSortProcess(Est_String ime);

#ifdef __cplusplus
}
#endif

#endif  /* _PA */

